extern void help_initialize(void);
extern int help_prepare(void);
extern void cx_help_link(void);
extern void cx_help_back(void);
extern void cx_help_main(void);
extern void cx_help_up(void);
extern void cx_help_down(void);
extern void cx_help_mouse(void);

#define MAIN_LINKS	5	/* max number of generated links in the MAIN page */
