extern void cx_pan_down(void);
extern void pan_down_n(int);
extern void cx_pan_up(void);
extern void pan_up_n(int);
extern void pan_line(int);
extern void cx_pan_end(void);
extern void cx_pan_home(void);
extern void cx_pan_pgdown(void);
extern void cx_pan_pgup(void);
extern void cx_pan_middle(void);
extern void cx_pan_mouse(void);
extern void pan_adjust(PANEL_DESC *);
