extern int fopt_prepare(void);
extern const char *fopt_saveopt(void);
extern int fopt_restoreopt(const char *);
extern void cx_fopt_enter(void);
extern void cx_filteredit_begin(void);
extern void cx_filteredit_end(void);
extern void cx_filteredit_left(void);
extern void cx_filteredit_right(void);
extern void cx_filteredit_kill(void);
extern void cx_filteredit_backsp(void);
extern void cx_filteredit_delchar(void);
extern void cx_filteredit_delend(void);
extern void filteredit_insertchar(wchar_t);
extern void filteredit_nu_insertchar(wchar_t);
extern void filter_update(void);
extern void filter_off(void);
extern void filter_help(void);
extern void cx_filter(void);
extern void cx_filter2(void);
