extern int sort_prepare(void);
extern const char *sort_saveopt(void);
extern int sort_restoreopt(const char *);
extern void sort_files(void);
extern void cx_sort_set(void);
extern int num_wcscoll(const wchar_t *, const wchar_t *);
