extern void match_pattern_set(const wchar_t *);
extern int match_pattern(const char *);
extern void match_substr_set(const wchar_t *);
extern int match_substr(const wchar_t *);
extern int match_substr_ic(const wchar_t *);
