extern void control_loop(int);
extern const char **mode2help(int);
extern int get_current_mode(void);
extern int get_previous_mode(void);
extern void fopt_change(void);
extern void cx_version(void);
extern void err_exit(const char *, ...);
